import os
import re
from dataclasses import dataclass, field
from subprocess import CalledProcessError, run as sub_run
from typing import Optional, List, Union

from pyterraformer.constants import logger
from pyterraformer.settings import get_default_terraform_location
from pyterraformer.terraform.backends import BaseBackend, LocalBackend


@dataclass
class Terraform:
    terraform_exec_path: Optional[str] = field(
        default_factory=get_default_terraform_location
    )
    plugin_cache_directory: Optional[str] = None
    backend: BaseBackend = field(default_factory=lambda: LocalBackend(path=os.getcwd()))
    workspace: str = "default"

    def run(self, arguments: Union[str, List[str]], path: str):
        workspace = self._run(["workspace", "show"], path=path).strip()
        logger.info(f"Executing {arguments} in workspace {workspace}.")
        if workspace != self.workspace:
            logger.info(f"swapping to configured workspace {workspace}")
            workspaces = [
                v.replace("*", "").strip()
                for v in self._run(["workspace", "list"], path=path).split("\n")
            ]
            if self.workspace in workspaces:
                logger.info("workspace found, swapping to")
                self._run(["workspace", "select", self.workspace], path=path)
            else:
                logger.info("workspace not found, creating")
                self._run(["workspace", "new", self.workspace], path=path)
        self._run(["init"], path=path)
        return self._run(arguments=arguments, path=path)

    def _run(self, arguments: Union[str, List[str]], path: str):
        if not self.terraform_exec_path:
            raise ValueError("No terraform executable set, cannot run TF commands.")
        runtime_env = os.environ.copy()
        if isinstance(arguments, str):
            arguments = [arguments]
        # os.cwd will fail on Null values, which can happen
        runtime_env = {
            key: value
            for key, value in {
                **runtime_env,
                **self.backend.generate_environment(),
            }.items()
            if value
        }
        if self.plugin_cache_directory:
            runtime_env["TF_PLUGIN_CACHE_DIR"] = self.plugin_cache_directory

        cmd_array: List[str] = [self.terraform_exec_path, *arguments]

        def run_cmd():
            return sub_run(
                cmd_array,
                cwd=path,
                env=runtime_env,
                check=True,
                capture_output=True,
                encoding="utf-8",
            ).stdout

        try:
            return run_cmd()
        except CalledProcessError as e:
            logger.error(e.stderr)
            raise e


def extract_errors(input: str):
    found = re.findall(
        r"(Error:.*)(?:Error:|$)", input, re.IGNORECASE | re.MULTILINE | re.DOTALL
    )
    return "\n".join(found).strip()



from typing import Optional

from pyterraformer.core.objects import TerraformObject, ObjectMetadata


class Comment(TerraformObject):
    def __init__(
        self,
        text: str,
        multiline: bool = False,
        _metadata: Optional[ObjectMetadata] = None,
    ):
        self.multiline = multiline
        if self.multiline:
            text = f"/*{text}*/"
        else:
            text = text.strip()
        TerraformObject.__init__(
            self, "comment", tf_id=None, text=text, _metadata=_metadata
        )

    @property
    def has_content(self):
        return bool(self.text.strip())



from os import environ
from typing import Optional


def get_default_terraform_location() -> Optional[str]:
    """Attempt to discover default terraform location"""
    declared_path = environ.get("TERRAFORM_EXEC", None)
    if declared_path:
        return declared_path
    from platform import system
    from subprocess import CalledProcessError, run

    if system() == "Windows":
        cmd = ["where", "terraform"]
    else:
        cmd = ["which", "terraform"]

    try:
        output = run(cmd, check=True, capture_output=True, encoding="utf-8")
        if output.stdout:
            # where may return multiple lines
            output_str = output.stdout.split("\n")[0].strip()
            return output_str
        return None
    except CalledProcessError:
        return None



from typing import Optional

from pyterraformer.core.objects import TerraformObject, ObjectMetadata


class TerraformConfig(TerraformObject):
    def __init__(self, _metadata: Optional[ObjectMetadata] = None, **kwargs):
        # self.backends = [obj for obj in attributes if isinstance(obj, Backend)]
        TerraformObject.__init__(self, "terraform", _metadata=_metadata, **kwargs)

    # def render(self, variables=None):
    #     from pyterraformer.core.generics import Block, Literal
    #
    #     output = {
    #         "required_version": self.required_version,
    #         "backend": Block([Literal(backend.render()) for backend in self.backends]),
    #     }
    #     final = process_attribute(output)
    #     return self.template.render(render_attributes=final)



from typing import Optional

from pyterraformer.core.objects import TerraformObject, ObjectMetadata


class ModuleObject(TerraformObject):
    def __init__(self, tf_id, _metadata: Optional[ObjectMetadata] = None, **kwargs):
        TerraformObject.__init__(self, _type="module", tf_id=tf_id, **kwargs)



from pyterraformer.exceptions import TerraformExecutionError
import pytest


def test_readme():
    from pyterraformer import HumanSerializer

    hs = HumanSerializer(terraform="/path/to/my/terraform/binary/does/not/exist")

    example_string: str = """resource "aws_s3_bucket" "b" {
      bucket = "my-tf-test-bucket"
      
      tags = {
    Name        = "My bucket"
        Environment = "Dev"
      }
    }"""

    # parse a string into a list of terraform objects
    namespace = hs.parse_string(example_string)

    # get the bucket from that list
    bucket = namespace[0]

    # modify the bucket
    bucket.tags["Environment"] = "Prod"
    bucket.bucket = "my-updated-bucket"

    # and write the modified namespace back
    # formatting requires a valid terraform binary to be provided
    with pytest.raises(TerraformExecutionError):
        updated = hs.render_object(bucket, format=True)



from dataclasses import dataclass
from typing import Optional, List, Dict, ClassVar

from pyterraformer.terraform.backends.base_backend import BaseBackend


@dataclass
class GCSBackend(BaseBackend):
    """Stores the state as an object in a configurable prefix in a pre-existing bucket on Google Cloud Storage (GCS). The bucket must exist prior to configuring the backend."""

    credentials: Optional[str] = None
    impersonate_service_account: Optional[str] = None
    impersonate_service_account_delegations: Optional[List[str]] = None
    access_token: Optional[str] = None
    prefix: Optional[str] = None
    encryption_key: Optional[str] = None

    SECRET_FIELDS: ClassVar = ["encryption_key", "access_token"]

    def generate_environment(self) -> Dict:
        output = {}
        if self.credentials:
            output["GOOGLE_BACKEND_CREDENTIALS"] = self.credentials
            output["GOOGLE_CREDENTIALS"] = self.credentials
        if self.impersonate_service_account:
            output[
                "GOOGLE_IMPERSONATE_SERVICE_ACCOUNT"
            ] = self.impersonate_service_account
        if self.encryption_key:
            output["GOOGLE_ENCRYPTION_KEY"] = self.encryption_key
        return output



from pyterraformer.core.objects import TerraformObject, ObjectMetadata
from typing import Optional

# from pyterraformer.core.resources import


class Backend(TerraformObject):
    def __init__(self, name: str, _metadata: Optional[ObjectMetadata] = None, **kwargs):
        TerraformObject.__init__(self, "backend", _metadata=_metadata, **kwargs)
        self.name = str(name).replace('"', "")

    # def render(self, variables=None):
    #     final = {}
    #     for key in sorted(self.render_variables.keys()):
    #         if key != "name":
    #             final[key] = self.render_variables[key]
    #     final = process_attribute(final)
    #     return self.template.render(name=self.name, render_attributes=final)



from pyterraformer.core.objects import TerraformObject
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyterraformer.core.generics import Literal


class Variable(TerraformObject):
    def __init__(self, text, name, attributes):
        self.name = str(name).replace('"', "")
        TerraformObject.__init__(self, "variable", text, attributes)

    def __repr__(self):
        return (
            f"{self._type}(name={self.name}, "
            + ", ".join(
                [f'{key}="{val}"' for key, val in self.render_variables.items()]
            )
            + ")"
        )

    def __getitem__(self, val):
        for key, item in self.default.items():

            if val == key:
                return item
        raise KeyError(val)

    def render_lookup(self, item) -> "Literal":
        from pyterraformer.core.generics import Literal

        return Literal(f'var.{self.name}["{item}"]')

    def render_attribute(self, item) -> "Literal":
        from pyterraformer.core.generics import Literal

        return Literal(f"var.{self.name}.{item}")

    def render_basic(self) -> "Literal":
        from pyterraformer.core.generics import Literal

        return Literal(f"var.{self.name}")

    def get(self, val, fallback: str = None):
        for key, item in self.default.items():
            if val == key:
                return item
        return fallback

    def get_type(self, val) -> "Literal":
        from pyterraformer.core.generics.interpolation import String
        from pyterraformer.core.generics import Literal, StringLit

        if isinstance(val, list):
            if not val:
                return Literal("list(any)")
            return Literal(f"list({self.get_type(val[0]).value})")
        elif isinstance(val, dict):
            out = set()
            for item in val.values():
                out.add(self.get_type(item).value)
            outlist = list(out)
            if len(outlist) == 1:
                return Literal(f"map({outlist[0]})")
            return Literal("map(any)")
        elif isinstance(val, set):
            return Literal(f"set({self.get_type(next(iter(val))).value})")
        elif isinstance(val, str):
            return Literal("string")
        elif isinstance(val, StringLit):
            return Literal("string")
        elif isinstance(val, bool):
            return Literal("bool")
        elif isinstance(val, int) or isinstance(val, float):
            return Literal("number")
        elif isinstance(val, String):
            return Literal("string")
        return Literal("any")

    def render(self, variables=None):
        variables = variables or {}
        variables["tf_id"] = self.name
        type_map = self.get_type(self.render_variables.get("default", None))
        self.render_variables["type"] = type_map

        priority_attributes = ["type"]

        # sort logic for workflow_utility attributes at top, then alphabetical
        final = {}
        for val in priority_attributes:
            test = self.render_variables.get(val)
            if test:
                final[val] = test
        for key, value in self.render_variables.items():
            if key in ("_type",):
                continue
            final[key] = value

        return self.template.render(render_attributes=final, **variables)



import os

import jinja2

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), "templates")
template_loader = jinja2.FileSystemLoader(searchpath=TEMPLATE_PATH)
env = jinja2.Environment(
    loader=template_loader, autoescape=True, keep_trailing_newline=True
)


def get_template(name: str):
    return env.get_template(f"{name}.tf")



from dataclasses import dataclass
from typing import Dict, ClassVar

from pyterraformer.core.generics import Backend


@dataclass
class BaseBackend:
    SECRET_FIELDS: ClassVar = []

    def generate_environment(self) -> Dict:
        return {}

    def as_object(self) -> Backend:
        return Backend("undefined")



#!python

import json
import os
from logging import DEBUG
from logging import StreamHandler
from pathlib import Path
from typing import Optional, List, Dict

from jinja2 import Template

from pyterraformer import HumanSerializer
from pyterraformer.constants import logger
from pyterraformer.core import TerraformWorkspace
from pyterraformer.terraform import Terraform
from pyterraformer.terraform.backends.local_backend import LocalBackend
import click

logger.addHandler(StreamHandler())
logger.setLevel(DEBUG)

TEMPLATE = Template(
    '''# This file is auto-generated. Do not edit manually.
{%- macro render_blocks(blocks, indent) -%}
{%- filter indent(width=indent) %}
{%- for key, block in blocks.items()  %}
{%- if block.nesting_mode == 'single' %}
@dataclass 
class {{block.camel_case}}():
    {{ render_blocks(block.blocks, 4) }}
    {%- for item in block.attributes %}{% if not item.optional and not item.computed %}
    {{item.name}}:{{ item.python_type }}{% endif %}{% endfor %}
    # non-optional-blocks
    
    {%- for key, item in block.blocks.items() %}{% if not item.optional and not item.computed %}
    {{item.snake_case}}:{{item.camel_case}}{% endif %}{% endfor %}
    
    {%- for item in block.attributes %}{% if item.optional %}
    {{item.name}}: {{item.python_type}} = EMPTY_DEFAULT{% endif %}{% endfor %}
    
    {%- for key, item  in block.blocks.items() %}{% if item.optional %}
    {{item.snake_case}}: Optional[{{item.camel_case}}]=EMPTY_DEFAULT{% endif %}{% endfor %}
    pass
{% endif %}
{%- if block.nesting_mode == 'list' %}
# wrapper list class
class {{block.camel_case}}(BlockList):
    @dataclass 
    class {{block.camel_case}}Item():
{{ render_blocks(block.blocks, 8) }}
        {%- for item in block.attributes %}{% if not item.optional and not item.computed %}
        {{item.name}}:{{ item.python_type }}{% endif %}{% endfor %}
        # non-optional-blocks
        
        {%- for key, item in block.blocks.items() %}{% if not item.optional and not item.computed %}
        {{item.snake_case}}:{{item.camel_case}}{% endif %}{% endfor %}
        
        {%- for item in block.attributes %}{% if item.optional %}
        {{item.name}}: Optional[{{item.python_type}}] = EMPTY_DEFAULT{% endif %}{% endfor %}
        
        {%- for key, item  in block.blocks.items() %}{% if item.optional %}
        {{item.snake_case}}: Optional[{{item.camel_case}}]=EMPTY_DEFAULT{% endif %}{% endfor %}
        pass
    items: List[{{block.camel_case}}Item]
{% endif %}
{% if block.nesting_mode == 'set' %}      
# wrapper set class
class {{block.camel_case}}(BlockSet):
    @dataclass 
    class {{block.camel_case}}Item():
{{ render_blocks(block.blocks, 8) }}
        {%- for item in block.attributes %}{% if not item.optional and not item.computed %}
        {{item.name}}:{{ item.python_type }}{% endif %}{% endfor %}
        # non-optional-blocks
        
        {%- for key, item in block.blocks.items() %}{% if not item.optional and not item.computed %}
        {{item.snake_case}}:{{item.camel_case}}{% endif %}{% endfor %}
        
        {%- for item in block.attributes %}{% if item.optional %}
        {{item.name}}: Optional[{{item.python_type}}] = EMPTY_DEFAULT{% endif %}{% endfor %}
        
        {%- for key, item  in block.blocks.items() %}{% if item.optional %}
        {{item.snake_case}}: Optional[{{item.camel_case}}]=EMPTY_DEFAULT{% endif %}{% endfor %}
        pass
    items: Set[{{block.camel_case}}Item]
{% endif %}
{% endfor %}
{% endfilter %}
{%- endmacro %}

from typing import List, Optional, Dict, Set, Any
from pyterraformer.core.resources import ResourceObject
from pyterraformer.core.generics import BlockList, BlockSet
from pyterraformer.core.objects import ObjectMetadata
from pyterraformer.constants import EMPTY_DEFAULT
from dataclasses import dataclass



class {{resource.camel_case}}(ResourceObject):
    """    
    Args:
    {%- for item in resource.attributes %}{% if not item.computed and not item.optional %}
        {{item.name}} ({{ item.python_type }}): {% filter indent(width=20) %}{{item.description}}{% endfilter %}{% endif %}{% endfor %}
        {%- for key, item  in blocks.items() %}{% if  not item.computed and not item.optional %}
        {{item.snake_case}}: Optional[{{item.camel_case}}]{% endif %}{% endfor %}
    """
    _type = '{{resource.snake_case}}'
    
{{ render_blocks(blocks, 4) }}
    
    def __init__(self,
        tf_id: str,
        
        {%- for item in resource.attributes %}{% if not item.optional and not item.computed %}
        {{item.name}}:{{ item.python_type }},{% endif %}{% endfor %}
        # non-optional-blocks
        
        {%- for key, item  in blocks.items() %}{% if not item.optional and not item.computed %}
        {{item.snake_case}}:{{item.camel_case}},{% endif %}{% endfor %}
        #optional
        _metadata: Optional[ObjectMetadata] = EMPTY_DEFAULT,
        
        {%- for item in resource.attributes %}{% if item.optional %}
        {{item.name}}: {{item.python_type}} = EMPTY_DEFAULT,{% endif %}{% endfor %}
        
        {%- for key, item  in blocks.items() %}{% if item.optional %}
        {{item.snake_case}}: Optional[{{item.camel_case}}]=None,{% endif %}{% endfor %}
        ):
            kwargs = {}
            {% for item in resource.attributes if not item.computed or item.optional %}if {{item.name}} is not None:
                kwargs['{{item.name}}'] = {{ item.name }}
            {% endfor %}
            {%- for key, item in blocks.items() if not item.computed or item.optional %}if {{item.snake_case}} is not None:
                kwargs['{{item.snake_case}}'] = {{ item.snake_case }}
            {% endfor %}
            super().__init__(tf_id=tf_id, _metadata=_metadata, **kwargs)
            
        
        

'''
)
from dataclasses import dataclass, field


def camel_case(string: str) -> str:
    return "".join([v.capitalize() for v in string.split("_")])


def python_type(tf_type: str) -> str:
    if tf_type == "string":
        return "str"
    elif tf_type == "number":
        return "float"
    elif "map" in tf_type:
        type = python_type([v for v in tf_type if v != "map"][0])
        return f"Dict[str, {type}]"
    elif "list" in tf_type:
        type = python_type([v for v in tf_type if v != "list"][0])
        return f"List[{type}]"
    elif "set" in tf_type:
        type = python_type([v for v in tf_type if v != "set"][0])
        return f"Set[{type}]"
    elif "object" in tf_type:
        components = python_type([v for v in tf_type if v != "object"][0])
        return "Dict[str,Any]"
    return tf_type


@dataclass
class Attribute:
    name: str
    type: str
    optional: bool
    computed: bool
    description: Optional[str] = None
    default: Optional[str] = None

    @property
    def python_type(self) -> str:
        if self.optional:
            return f"Optional[{python_type(self.type)}]"
        return python_type(self.type)


@dataclass
class Resource:
    snake_case: str
    description: Optional[str]
    attributes: List[Attribute]

    @property
    def camel_case(self):
        return "".join([v.capitalize() for v in self.snake_case.split("_")])

    @classmethod
    def parse_from_resource_info(cls, key: str, info: dict):
        attributes = [
            Attribute(
                name=key,
                type=v.get("type"),
                optional=v.get("optional"),
                computed=v.get("computed", False),
                default=v.get("default", None),
                description=v.get("description", ""),
            )
            for key, v in info["block"]["attributes"].items()
        ]
        # move optional attributes to the end
        attributes = sorted(
            attributes, key=lambda x: "zzz" + x.name if x.optional else x.name
        )
        return Resource(key, description=info.get("description"), attributes=attributes)


def generate_blocks(schema: dict, depth=0) -> dict:
    # extra_blocks = {}
    # for key, value in schema['block'].get('block_types', {}).items():
    #
    #     extra_blocks = {**generate_blocks(value, depth + 1), **extra_blocks}
    blocks = {
        key: Block(
            snake_case=key,
            nesting_mode=value.get("nesting_mode"),
            description=value.get("description", ""),
            max_items=value.get("max_items"),
            blocks=generate_blocks(value),
            optional=value.get("optional", False),
            attributes=[
                Attribute(
                    name=key,
                    type=v.get("type"),
                    optional=v.get("optional"),
                    computed=v.get("computed"),
                    default=v.get("default"),
                    description=v.get("description", ""),
                )
                for key, v in value["block"].get("attributes", {}).items()
            ],
        )
        for key, value in schema["block"].get("block_types", {}).items()
    }
    return blocks


@dataclass
class Block:
    snake_case: str
    optional: bool
    attributes: List[Attribute]
    description: Optional[str]
    nesting_mode: str
    max_items: int = -1
    blocks: [Optional[Dict[str, "Block"]]] = field(default_factory=dict)

    def __post_init__(self):
        # all lists can be empty
        # single maybe shouldn't be the case
        # but for now
        if self.nesting_mode in ("list", "set", "single"):
            self.optional = True

    @property
    def camel_case(self):
        return "".join([v.capitalize() for v in self.snake_case.split("_")])


def build_resource_class(key: str, schema: dict):
    blocks = generate_blocks(schema)
    # blocks = sorted(blocks, key= lambda x: not x.blocks)
    class_text = TEMPLATE.render(
        resource=Resource.parse_from_resource_info(key, schema), blocks=blocks
    )
    return class_text


def _create_stubs(provider: str, version: str, output_path: str, terraform_path: str):
    output = {}
    tf = Terraform(
        terraform_exec_path=terraform_path, backend=LocalBackend(path=os.getcwd())
    )
    workspace = TerraformWorkspace(
        terraform=tf, path=os.getcwd(), serializer=HumanSerializer(terraform=tf)
    )
    workspace.add_provider(name="build", source=provider, version=version)
    workspace.save_all()
    tf.run("init", path=os.getcwd())
    provider_info = tf.run(["providers", "schema", "-json"], path=os.getcwd())
    loaded = json.loads(provider_info)
    schemas = loaded["provider_schemas"]
    google = schemas[f"registry.terraform.io/{provider}"]
    resources = google["resource_schemas"]
    for key, value in resources.items():
        if key == "google_bigquery_connection":
            print(key)
            print(value)
        output[key] = build_resource_class(key, value)

    save_stubs(provider, version, output, output_path)


def save_stubs(provider: str, version: str, stubs: dict, path: str):
    base_path = Path(path)
    root = Path(path) / provider / f"v{version.replace('.','_')}"
    os.makedirs(root, exist_ok=True)
    for key, text in stubs.items():
        if not text:
            continue
        with open(root / f"{key}.py", "w", encoding="utf-8") as f:
            f.write(text)
    imports = [f"from .{v} import {camel_case(v)}" for v in stubs.keys()]
    all = ",".join([f'"{camel_case(v)}"' for v in stubs.keys()])
    with open(root / "__init__.py", "w", encoding="utf-8") as f:
        f.write("\n".join(imports))
        f.write(f"\n__all__ = [{all}]")
    parts = root.relative_to(base_path).parts
    for idx in range(1, len(parts)):
        subpath = Path(path)
        for val in parts[0:idx]:
            subpath = subpath / val
        with open(subpath / "__init__.py", "w", encoding="utf-8") as f:
            f.write("\n")


@click.command()
@click.option(
    "--provider", help="The provider name, eg. hashicorp/aws to create stubs for"
)
@click.option("--version", help="The version of the provider to create stubs for. ")
@click.option("--terraform_path", help="The path of the local terraform binary.")
@click.option("--output_path", help="The output path to put the created files in")
def create_stubs(provider: str, version: str, output_path: str, terraform_path: str):
    _create_stubs(
        provider=provider,
        version=version,
        output_path=output_path,
        terraform_path=terraform_path,
    )


if __name__ == "__main__":
    create_stubs()

    # tf = Terraform(terraform_exec_path=r'c:/tech/terraform.exe',
    #                backend=LocalBackend(path=os.getcwd()), )
    # workspace = TerraformWorkspace(terraform=tf, path=os.getcwd(), serializer=HumanSerializer(terraform=tf))
    # workspace.add_provider(name='google', source="hashicorp/google")
    # workspace.save_all()
    # tf.run('init', path=os.getcwd())
    # provider_info = tf.run(['providers', 'schema', '-json'], path=os.getcwd())
    # loaded = json.loads(provider_info)
    # schemas = loaded['provider_schemas']
    # google = schemas['registry.terraform.io/hashicorp/google']
    # resources = google['resource_schemas']
    # for key, value in resources.items():
    #     if 'google_storage_bucket'==key:
    #         build_resource_class(key, value)



from types import MethodType


class lazy_property(object):
    """ used for lazy evaluation of an object attribute. property should represent non-mutable data, as it replaces itself."""

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)

        setattr(obj, "_method_{}".format(self.func_name), MethodType(self.fget, obj))
        return value



from pyterraformer.core.objects import TerraformObject


class Metadata(TerraformObject):
    def __init__(self, text, attributes):
        TerraformObject.__init__(self, "metadata", text, attributes)



#!python
import requests

try:
    from BeautifulSoup import BeautifulSoup
except ImportError:
    from bs4 import BeautifulSoup

if __name__ == "__main__":
    providers = requests.get("https://releases.hashicorp.com/")
    provider_set = set()
    parsed_html = BeautifulSoup(providers.text)
    for match in parsed_html.body.find_all("a"):
        if match.text.startswith("terraform-provider"):
            provider_set.add(match.text.replace("terraform-provider-", ""))
    for provider in provider_set:
        version_set = set()
        if provider != "google":
            continue
        versions = requests.get(
            f"https://releases.hashicorp.com/terraform-provider-{provider}/"
        )
        versions_html = BeautifulSoup(versions.text)
        for match in versions_html.body.find_all("a"):
            if "_" in match.text:
                # terraform-provider-google_4.27.0
                version = match.text.rsplit("_")[1]
                version_set.add(version)

        print(version_set)



from typing import Optional, Dict, TYPE_CHECKING, Union
from pathlib import Path

if TYPE_CHECKING:
    from pyterraformer.core import (
        TerraformWorkspace,
        TerraformObject,
        TerraformNamespace,
    )


class BaseSerializer(object):
    pass

    @property
    def can_format(self) -> bool:
        return False

    def parse_string(self, string: str):
        raise NotImplementedError

    def parse_file(self, path: Union[str, Path], workspace: "TerraformWorkspace"):
        raise NotImplementedError

    def _format_string(self, string: str):
        raise NotImplementedError

    def render_object(
        self, object: "TerraformObject", format: Optional[bool] = None
    ) -> str:
        raise NotImplementedError

    def render_namespace(
        self, namespace: "TerraformNamespace", format: Optional[bool] = None
    ) -> str:
        raise NotImplementedError

    def render_workspace(self, workspace: "TerraformWorkspace") -> Dict[str, str]:
        raise NotImplementedError

    #



from dataclasses import dataclass
from typing import Dict, Optional, TYPE_CHECKING

from pyterraformer.exceptions import ValidationError


if TYPE_CHECKING:
    from pyterraformer.core.namespace import TerraformNamespace


@dataclass
class ObjectMetadata:
    source_file: Optional[str] = None
    orig_text: Optional[str] = None
    row_num: Optional[int] = None
    start_pos: Optional[int] = None
    end_pos: Optional[int] = None


class TerraformObject(object):
    def __init__(
        self,
        _type,
        tf_id: Optional[str] = None,
        _metadata: Optional[ObjectMetadata] = None,
        **kwargs,
    ):

        self._metadata = _metadata or ObjectMetadata()
        self.tf_id = tf_id
        arguments = kwargs or {}
        self.render_variables: Dict[str, str] = {
            str(key): value for key, value in arguments.items()
        }
        # for attribute in self.attributes:
        #     if isinstance(attribute, list):
        #         # always cast keys to string
        #         self.render_variables[str(attribute[0])] = attribute[1]
        #     elif isinstance(attribute, Block):
        #         base = self.render_variables.get(attribute.name, Block())
        #         base.append(attribute)
        #         self.render_variables[attribute.name] = base
        self._type: str = _type
        self._changed: bool = False
        self._workspace = None
        self._file: Optional["TerraformNamespace"] = None
        self._initialized: bool = True

    def __repr__(self):
        return (
            f"{self._type}("
            + ", ".join(
                [f'{key}="{val}"' for key, val in self.render_variables.items()]
            )
            + ")"
        )

    @property
    def tf_attributes(self):
        return self.render_variables.keys()

    def __getattr__(self, name):
        if name == "render_variables":
            return self.__dict__.get("render_variables")
        elif self.render_variables and name in self.render_variables:
            return self.__dict__.get("render_variables").get(name)
        elif (
            self.render_variables
            and "_" in name
            and name.endswith("resolved")
            and name.rsplit("_", 1)[0] in self.render_variables
        ):
            lookup = name.rsplit("_", 1)[0]
            return self.resolved_attributes.get(lookup)
        elif "_" in name and name.rsplit("_", 1)[1] == "lookup":
            from pyterraformer.core.generics import Literal

            property = name.rsplit("_", 1)[0]
            return Literal(f"{self._type}.{self.tf_id}.{property}")
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        """This gets tricky:
        We want any new attribute set on a base class to get rendered...
        Unless it's one of the internal attributes we use for managing objects.
        So skip anything with a private method, or in the disallow list...
        Unless it's also in the list of things that we should render."""
        if name.startswith("_") or (
            name in ("row_num", "template", "name", "tf_id")
            and not (self.render_variables and name in self.render_variables)
        ):
            super().__setattr__(name, value)
        elif (self.render_variables and name in self.render_variables) or getattr(
            self, "_initialized", False
        ):
            self._changed = True
            self.__dict__.get("render_variables")[name] = value
        else:
            super().__setattr__(name, value)

    def __delattr__(self, name):
        if self.render_variables and name in self.render_variables:
            self._changed = True
            del self.__dict__.get("render_variables")[name]
        else:
            super().__delattr__(name)

    @property
    def changed(self):
        return self._original_text and not self._changed

    def resolve_item(self, item):
        from pyterraformer.core.generics import Variable

        if isinstance(item, list):
            resolved = [self.resolve_item(sub_item) for sub_item in item]
        elif isinstance(item, str):
            resolved = item
        elif isinstance(item, int):
            resolved = item
        elif isinstance(item, dict):
            resolved = {
                self.resolve_item(key): self.resolve_item(value)
                for key, value in item.items()
            }
        else:
            resolved = item.resolve(self._workspace, self._file, None, None)
        if isinstance(resolved, Variable) and hasattr(resolved, "default"):
            resolved = getattr(resolved, "default")
        return resolved

    @property
    def resolved_attributes(self):
        out = {}
        for key, item in self.render_variables.items():
            resolved = self.resolve_item(item)
            out[str(key)] = resolved

        return out



from typing import Dict, Tuple, Any

from lark import Lark, Transformer, v_args
from lark.tree import Meta

from pyterraformer.core.generics import (
    Comment,
    Variable,
    Interpolation,
    String,
    StringLit,
    DictLookup,
    PropertyLookup,
    TerraformConfig,
    Provider,
    BinaryTerm,
    Data,
    Expression,
    Conditional,
    BinaryOp,
    BinaryOperator,
    Output,
    Local,
    Parenthetical,
    LegacySplat,
    BlockList,
    File,
    Boolean,
    Merge,
    Concat,
    Types,
    Replace,
    ArrayLookup,
    GenericFunction,
    Symlink,
    ToSet,
)
from pyterraformer.core.modules import ModuleObject
from pyterraformer.core.objects import ObjectMetadata, TerraformObject
from typing import List

# TODO: rewrite to comply with https://github.com/hashicorp/hcl2/blob/master/hcl/hclsyntax/spec.md

RESOURCES_MAP: Dict = {}

grammar = r"""
    start: ( item | symlink)*
    ?item: resource
    | object
    | module
    | provider
    | terraform
    | variable
    | locals
    | data
    | output
    | multiline_comment
    | comment

    //TODO: in parsing - follow symlinks?
    symlink: "."+ "/" /[a-zA-Z_\-]+/ ".tf"

    resource: "resource" string_lit string_lit  "{" (nested_comment| sub_object | lifecycle | provider | split_subarray ) + "}"
    
    nested_comment: comment
    
    output: "output" string_lit "{"[( nested_comment| sub_object)+] "}"

    module: "module" string_lit  "{"  (nested_comment | sub_object | split_subarray)+ "}"

    lifecycle: "lifecycle"  "{" [sub_object ( sub_object)*] "}"

    // this exists without the = option as a perf optimization
    split_subarray: IDENTIFIER "{" [(nested_comment|sub_object|split_subarray)+] "}"

    provider: "provider" (string_lit | IDENTIFIER)  "{" [(sub_object | split_subarray) ( sub_object| split_subarray)*]  "}"

    data: "data" string_lit string_lit "{" [(sub_object | split_subarray | nested_comment) ( sub_object| split_subarray | nested_comment)*] "}"

    metadata: "metadata" "{" [sub_object ( sub_object)*] "}"

    object: "object" string_lit "{" [sub_object ( sub_object)*] "}"

    locals: "locals" dict

    variable: "variable" (string_lit | IDENTIFIER) "{" (nested_comment | variable_type_declaration| variable_default_declaration | variable_description)* "}"

    variable_type_declaration: ("type" | "\"type\"") "=" (types | ( "\""  TYPE"\"" ))
    variable_description: ("description" | "\"description\"") "=" (string_lit | heredoc_eof)
    variable_default_declaration: ("default"| "\"default\"") "=" (tuple | string_lit | boolean | dict | int_lit | heredoc_eof)

    // terraform specific settings
    terraform: "terraform" "{" (backend | sub_object | split_subarray )* "}"

    backend: "backend" string_lit  "{" [sub_object ( sub_object)*] "}"

    // handling inside items
    // figure out how to handle comments better
    tuple: "["  [ nested_comment* (string_lit| int_lit| dict| tuple | lookup | EMPTY_STRING )  ( "," nested_comment*  (string_lit| int_lit| dict| tuple | lookup) )*] ","? nested_comment?  "]" -> tuple

    // replace with expr_contents
    // figure out why dot identifiers are ever valid
    sub_object: (IDENTIFIER |string_lit | IDENTIFIER_WITH_DOT) "=" ( setsubtract | lookup | tuple | string_lit | file | toset | concat | merge | boolean | dict | int_lit | object_access |  conditional | heredoc_eof | IDENTIFIER | generic_function | list_comp) ","? nested_comment?    -> sub_object

    // annoyingly repetitive
    dict_sub_object: (IDENTIFIER |string_lit | IDENTIFIER_WITH_DOT) ("=" | ":") ( setsubtract | lookup | tuple | string_lit | file | toset | concat | merge | boolean | dict | int_lit | object_access |  conditional | heredoc_eof | IDENTIFIER | generic_function | list_comp) ","? nested_comment?  -> sub_object

    dict: "{" (nested_comment | dict_sub_object )* "}"

    object_access: IDENTIFIER "." IDENTIFIER

    null: "null"

    IDENTIFIER_WITH_DOT : /[a-zA-Z_][a-zA-Z0-9_\\-\\.]*/
    STRING_TERM : STRING_CHARS _WHITESPACE
        IDENTIFIER : /[a-zA-Z_][a-zA-Z0-9_-]*/
    //string section
    string_lit: "\"" ( STRING_CHARS | interpolation)* "\""
    STRING_CHARS: /(?:(?!\${)([^"\\]|\\.))+/+ // any character except '"" unless inside a interpolation string

    EMPTY_STRING: "\"" "\""
    interpolation: "${" (expr_contents)+ "}"
    // simple_interpolation:  lookup (expr_contents)+
    lookup  : IDENTIFIER "." (IDENTIFIER | lookup | dict_lookup | array_lookup | legacy_splat)+
    // interpolation to prioritize match
    // TODO: determine better way
    dict_lookup: IDENTIFIER "[" ( string_lit | lookup | interpolation | conditional | dict_lookup | concat | toset | setsubtract| merge | replace | )  "]"
    array_lookup: IDENTIFIER "[" DECIMAL "]"
    file: "file" "(" string_lit ","? ")"
    replace: "replace" "(" (expr_contents)+ "," string_lit "," string_lit ")"
    concat: "concat" "("  (expr_contents | tuple )+  (","  (expr_contents | tuple))* ","? ")"
    merge: "merge" "("  (expr_contents | tuple )+  ","  (expr_contents | tuple)+  ")"
    coalesce: "coalesce" "(" (expr_contents | tuple )  (  "," expr_contents | tuple)+   ")"
    contains: "contains" "(" (expr_contents | tuple )  (  "," expr_contents | tuple)+   ")"
    flatten: "flatten" "(" tuple ")"
    toset: "toset" "(" expr_contents ")"
    setsubtract: "setsubtract" "(" (expr_contents | tuple) "," (expr_contents | tuple) ")"
    list_lit: "list" "("  ( (EMPTY_STRING)| expr_contents )+ ")"
    element: "element" "(" (expr_contents)+ "," int_lit ")"
    generic_function: IDENTIFIER "(" (expr_contents)+ ")"

    ?expr_contents : conditional
                | interpolation
                | lookup
                | array_lookup
                | dict_lookup
                | IDENTIFIER
                | boolean
                | int_lit
                | string_lit
                | file
                | replace
                | flatten
                | toset
                | coalesce
                | contains
                | setsubtract
                | concat
                | merge
                | element
                | generic_function
                | binary_operator
                | parenthetical
                | legacy_splat
                | list_lit
                | heredoc_eof
                | heredoc_template
                | heredoc_template_trim
                | list_comp
                | nested_comment
                | tuple
                | null

    heredoc_eof: /<<-?(?P<delimiter>[^\s]+).*?(?P=delimiter)/s
    //heredoc_eof : ("<<EOF" | "<<-EOF" | "<<-EOT" | "<<EOT" | "<<-EOS" ) /((.|\n)(?!(EOF)|(EOT)|(EOS)))+/  /(.|\n){1}/ ("EOF" | "EOT" | "EOS" )
    // : ("<<" "-"? "EO" ("T" | "S" | "F"  ) ) /((.|\n)(?!(EOF)|(EOT)))+/  /(.|\n){1}/ ("EOF" | "EOT" | "EOS")
    heredoc_template : /<<(?P<heredoc>[a-zA-Z][a-zA-Z0-9._-]+)\n(?:.|\n)+?\n+\s*(?P=heredoc)/
    heredoc_template_trim : /<<-(?P<heredoc_trim>[a-zA-Z][a-zA-Z0-9._-]+)\n(?:.|\n)+?\n+\s*(?P=heredoc_trim)/

    expression : (expr_contents)+ | operation | conditional

    // be smarter about handling this
    legacy_splat : IDENTIFIER "." "*" "."

    conditional : expression "?" expression ":" expression
    list_comp: "[" "for" expression  ("," expression)* "in" expression ":" expression "]"
    ?operation : unary_op | binary_op
    !unary_op : ("-" | "!") expr_contents
    binary_op : expression binary_term
    !binary_operator : "==" | "!=" | "<" | ">" | "<=" | ">=" | "-" | "*" | "/" | "%" | "&&" | "||" | "+" -> bool_token
    binary_term : binary_operator expression

    parenthetical : "(" (expr_contents)+ ")"

    !boolean : "true" | "false"
    DECIMAL: "0".."9"
    int_lit: "-"? DECIMAL+

    comment :   /#.*(\n|$)/ |  /\/\/.*\n/

    multiline_comment: "/*" /((.|\n)(?!\*\/))+/  /(.|\n)/ "*/"

    //variables
    // map is valid and refers to map(any) https://www.terraform.io/docs/configuration/types.html#map-
    TYPE : "string" | "number" | "bool" | "map" | "list" | "any"
    // TODO: finish object and tuple
    types:(("tuple" | "set" | "map" | "object" | "list" ) "(" types ")") | TYPE

    %import common.WS_INLINE -> _WHITESPACE
    %import common.WS
    %ignore WS
"""


def args_to_dict(input_list: list) -> Dict[str, Any]:
    output: Dict[str, Any] = {}
    for array in input_list:
        key = array[0]
        val = array[1]
        # a comment might show up as a third object on a line
        # flatten that out
        if len(array) > 2:
            nested = array[2]
            output = {**output, **args_to_dict([nested])}
        if key not in output:
            output[str(key)] = val
        else:
            output[key] += val
    return output


class ParseToObjects(Transformer):
    def __init__(self, visit_tokens, text):
        Transformer.__init__(self, visit_tokens)
        self.text = text

    def meta_to_text(self, meta: Meta):
        return self.text[meta.start_pos : meta.end_pos]

    def generate_metadata(self, meta: Meta) -> ObjectMetadata:
        return ObjectMetadata(
            orig_text=self.text[meta.start_pos : meta.end_pos],
            start_pos=meta.start_pos,
            end_pos=meta.end_pos,
            row_num=meta.line,
        )

    def IDENTIFIER(self, args):
        return String(args.value)

    def STRING_CHARS(self, args):
        return String(args.value)

    def string_lit(self, args):
        return StringLit(args)

    def DECIMAL(self, args):
        return int(args.value)

    @v_args(meta=True)
    def resource(self, meta: Meta, args):
        from pyterraformer.core.resources import ResourceObject

        metadata = self.generate_metadata(meta)
        _type, name = args[0:2]
        _type = str(_type).replace('"', "")
        object_type = RESOURCES_MAP.get(_type, ResourceObject)
        object_type._type = _type
        # out = RESOURCES_MAP[str(type).replace('"', "")](
        #     name, str(type), , args[2:]
        # )
        remaining = args[2:]
        parsed = args_to_dict(remaining)
        return object_type(tf_id=name, _metadata=metadata, **parsed)

    @v_args(meta=True)
    def module(self, meta: Meta, args):
        name = args[0]
        remaining = args[1:]
        parsed = args_to_dict(remaining)
        out = ModuleObject(tf_id=name, _metadata=self.meta_to_text(meta), **parsed)
        return out

    @v_args(meta=True)
    def variable(self, meta: Meta, args):
        name = args[0]
        args = args[1:]
        out = Variable(self.meta_to_text(meta), name, args)
        out.row_num = meta.start_pos
        return out

    @v_args(meta=True)
    def provider(self, meta: Meta, args):
        name = args[0]
        metadata = self.generate_metadata(meta)
        remaining = args[0:]
        parsed = args_to_dict(remaining)
        out = Provider(name, _metadata=metadata, **parsed)
        return out

    @v_args(meta=True)
    def metadata(self, meta: Meta, args):
        metadata = self.generate_metadata(meta)
        remaining = args[0:]
        parsed = args_to_dict(remaining)
        out = TerraformConfig(_metadata=metadata, **parsed)
        return out

    @v_args(meta=True)
    def terraform(self, meta: Meta, args):
        from pyterraformer.core.objects import ObjectMetadata

        metadata = ObjectMetadata(
            orig_text=self.meta_to_text(meta), row_num=meta.start_pos
        )

        parsed = args_to_dict(args)
        return TerraformConfig(_metadata=metadata, **parsed)

    @v_args(meta=True)
    def data(self, meta: Meta, args):
        type, name = args[0:2]
        out = Data(name, type, self.meta_to_text(meta), args[2:])
        out.row_num = meta.start_pos
        return out

    @v_args(meta=True)
    def locals(self, meta: Meta, args):
        out = Local(self.meta_to_text(meta), args[0])
        out.row_num = meta.start_pos
        return out

    @v_args(meta=True)
    def comment(self, meta: Meta, args):
        metadata = self.generate_metadata(meta)
        out = Comment(text=args[0].value, _metadata=metadata)
        return out

    @v_args(meta=True)
    def nested_comment(self, meta: Meta, args):
        """Special comment to maintain position within other objects"""
        return [f"comment-{meta.start_pos}", args[0]]

    @v_args(meta=True)
    def multiline_comment(self, meta: Meta, args):
        metadata = ObjectMetadata(
            orig_text=self.meta_to_text(meta), row_num=meta.start_pos
        )
        base = args[0].value
        if len(args) > 1:
            base += args[1].value
        out = Comment(text=base, multiline=True, _metadata=metadata)
        return out

    @v_args(meta=True)
    def backend(self, meta: Meta, args):
        from pyterraformer.core.generics import Backend

        metadata = ObjectMetadata(
            orig_text=self.meta_to_text(meta), row_num=meta.start_pos
        )
        return "backend", Backend(args[0], _metadata=metadata)

    def output(self, args):
        return Output(args[0], args[1:])

    def tuple(self, args):
        return [*args]

    def sub_object(self, args):
        return args

    def variable_type_declaration(self, args):
        return ["_type", args[0]]

    def variable_default_declaration(self, args):
        return ["default", args[0]]

    def dict(self, args):
        return {key[0]: key[1] for key in args if not isinstance(key, Comment)}

    def interpolation(self, args):
        return Interpolation(args)

    def parenthetical(self, args):
        return Parenthetical(args)

    def legacy_splat(self, args):
        return LegacySplat(args)

    def start(self, args):
        return list(args)

    def lookup(self, args):
        base = args[0]
        return PropertyLookup(base, args[1:])

    def dict_lookup(self, args):
        return DictLookup(args[0], args[1:])

    def array_lookup(self, args):
        return ArrayLookup(args[0], args[1:])

    def expression(self, args):
        return Expression(args)

    def conditional(self, args):
        return Conditional(args)

    def binary_op(self, args):
        return BinaryOp(args)

    def binary_operator(self, args):
        return BinaryOperator(args)

    def file(self, args):
        return File(args)

    def types(self, args):
        return Types(args)

    def concat(self, args):
        return Concat(args)

    def merge(self, args):
        return Merge(args)

    def toset(self, args):
        return ToSet(args)

    def generic_function(self, args):
        return GenericFunction(args)

    def replace(self, args):
        return Replace(args)

    def int_lit(self, args):
        return int("".join([str(item) for item in args]))

    def binary_term(self, args):
        return BinaryTerm(args)

    def bool_token(self, args):
        return str(args[0].value)

    def split_subarray(self, args: list) -> Tuple[str, BlockList]:
        name = args[0]
        return (
            name,
            BlockList([{key: val} for key, val in args_to_dict(args[1:]).items()]),
        )

    def boolean(self, args):
        type = args[0]
        return Boolean(type == "true")

    def symlink(self, args):
        return Symlink(args)


TERRAFORM_PARSER = Lark(grammar, start="start", propagate_positions=True)


def parse_text(text: str) -> List[TerraformObject]:
    return ParseToObjects(visit_tokens=True, text=text).transform(
        TERRAFORM_PARSER.parse(text)
    )



from pyterraformer.core.objects import TerraformObject  # , process_attribute


class Local(TerraformObject):
    def __init__(self, text, attributes: dict):
        pass_on = []
        for key, value in attributes.items():
            pass_on.append([key, value])

        TerraformObject.__init__(
            self, _type="local", original_text=text, attributes=pass_on
        )

    # def render(self, variables=None):
    #     from analytics_terraformer_core.utility import clean_render_dictionary
    #
    #     variables = variables or {}
    #
    #     final = clean_render_dictionary(self.render_variables, [])
    #     output = process_attribute(final)
    #     return self.template.render(render_attributes=output, **variables)



import re
from collections import defaultdict
from fnmatch import fnmatch
from os.path import dirname
from pathlib import Path, PurePath
from typing import Dict, List, Union, Any
from typing import Optional, TYPE_CHECKING

from pyterraformer.constants import logger
from pyterraformer.core.generics import Literal, BlockList
from pyterraformer.core.utility import get_root
from pyterraformer.serializer import BaseSerializer
from pyterraformer.terraform import Terraform

if TYPE_CHECKING:
    from pyterraformer.core.namespace import TerraformFile
    from pyterraformer.core.generics.variables import Variable


def process_attribute(input: Any, level=0):
    from pyterraformer.core.generics import Variable

    # if isinstance(input, Block):
    #     input = {input.name: input._keys}

    if not isinstance(input, dict):
        return input
    output: Dict[str, Any] = {}
    for key, item in input.items():
        if isinstance(item, Variable):
            output[key] = item.render_basic()
        # elif isinstance(item, StringLit):
        #     output[key] = item.__repr__()
        elif isinstance(item, Literal):
            output[key] = item
        elif isinstance(item, BlockList):
            for idx, sub_item in enumerate(item):
                output[f"{key}~~block_{idx}"] = process_attribute(sub_item)
        elif isinstance(item, dict):
            output[key] = process_attribute(item)
        elif isinstance(item, List):
            output[key] = [process_attribute(sub_item) for sub_item in item]
        else:
            output[key] = item
    return output


class LazyFileDict(dict):
    def __init__(self, *args):
        dict.__init__(self, args)

    def iter(self):
        from pyterraformer.core.namespace import LazyFile

        for key, value in super().iter():
            if isinstance(value, LazyFile):
                value = value.resolve()
                self[key] = value
            yield key, value

    def items(self, resolve=True):
        from pyterraformer.core.namespace import LazyFile

        for key, value in super().items():
            if isinstance(value, LazyFile) and resolve:
                value = value.resolve()
                self[key] = value
            yield key, value

    def __getitem__(self, key):
        from pyterraformer.core.namespace import LazyFile

        val = dict.__getitem__(self, key)
        if isinstance(val, LazyFile):
            val = val.resolve()
            self[key] = val
        return val

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)


def extract_errors(input: str):
    found = re.findall(
        r"(Error:.*)(?:Error:|$)", input, re.IGNORECASE | re.MULTILINE | re.DOTALL
    )
    return "\n".join(found).strip()


class TerraformWorkspace(object):
    def __init__(
        self,
        path: Union[str, PurePath],
        terraform: Optional[Terraform] = None,
        serializer: Optional[BaseSerializer] = None,
        files: Optional[List["TerraformFile"]] = None,
        children: Optional[List["TerraformWorkspace"]] = None,
    ):

        self.terraform = terraform
        self.path = str(path)
        self._path = Path(self.path)
        self.files: Dict[str, TerraformFile] = LazyFileDict()
        if files:
            for file in files:
                self.files[file.name] = files
        self.children: List[TerraformWorkspace] = children or []
        self.name = self._path.stem
        self.variables: Dict[str, "Variable"] = {}
        self.data: List = []
        self.serializer = serializer

    def apply(self):
        return self.terraform.run(["apply", "--auto-approve"], path=self._path)

    @property
    def terraform_path(self):
        return get_root(self._path)

    def relative_path(self, path: str):
        return get_root(str(self._path), path)

    def get_file_safe(self, name: str) -> "TerraformFile":
        from pyterraformer.core.namespace import TerraformFile

        out = self.files.get(name)
        if not out:
            try:
                if not self.serializer:
                    raise FileNotFoundError("No valid serializer found.")
                file = self.serializer.parse_file(self._path / name, self)
            except FileNotFoundError:
                file = TerraformFile(
                    workspace=self, text="", location=self._path / name
                )

            self.files[name] = file

        return self.files[name]

    def get_terraform_config(self):
        from pyterraformer.core.generics import TerraformConfig
        from pyterraformer.core.generics import BlockList

        terraform = self.get_file_safe("terraform.tf")
        existing = [
            obj for obj in terraform.objects if isinstance(obj, TerraformConfig)
        ]
        if existing:
            return existing[0]
        logger.info("creating new terraform config")
        config = TerraformConfig(
            backend=self.terraform.backend.as_object(),
            required_providers=BlockList([{}]),
        )
        terraform.add_object(config)
        return config

    def add_provider(self, name: str, source: str, **kwargs):
        existing = self.get_terraform_config()
        required_providers = getattr(existing, "required_providers", BlockList([{}]))
        required_providers[0][name] = {"source": source, **kwargs}
        existing.required_providers = required_providers

    def add_file(self, file: Union["TerraformFile", PurePath, str]) -> "TerraformFile":
        from pyterraformer.core.namespace import TerraformFile

        if isinstance(file, TerraformFile):
            self.files[file.name] = file
            return file
        elif str(file) in self.files:
            return self.files[str(file)]
        else:
            nfile = TerraformFile(workspace=self, text="", location=self._path / file)
            self.files[nfile.name] = nfile
            return nfile

    def add_child_workspace(self, path: str):
        child = TerraformWorkspace(
            path=path, serializer=self.serializer, terraform=self.terraform
        )
        self.children.append(child)
        setattr(self, child.name, child)

    def add_variable(
        self, key: str, values, exists_okay: bool = False, replace: bool = False
    ):
        from pyterraformer.core.generics.variables import Variable

        if key in self.variables:
            if exists_okay:
                return self.variables[key]
            elif replace:
                pass
            else:
                raise ValueError(f"Key {key} already exists in variables!")

        if isinstance(values, dict):
            vtype = "map"
            variable = Variable(None, key, [["type", vtype], ["default", values]])
        elif isinstance(values, str):
            vtype = "string"
            variable = Variable(None, key, [["type", vtype], ["default", values]])
        elif isinstance(values, Variable):
            variable = values
        else:
            raise ValueError(f"Unable to map {type(values)} to a variable type")

        self.variables[key] = variable
        # create if not exists
        if "variables.tf" not in self.files:
            self.files["variables.tf"] = TerraformFile(
                self, "", self._path / "variables.tf"
            )

        self.files["variables.tf"].add_object(variable, replace=replace)
        return variable

    def delete_variable(self, key: str, safe=False):
        try:
            variable = self.variables[key]
        except KeyError as e:
            if safe:
                return
            raise e
        self.files["variables.tf"].delete_object(variable)

    def save(self, format: bool = True, apply: bool = False):
        from pyterraformer.core.namespace import LazyFile

        if not self.serializer:
            raise ValueError("Cannot save without serializer defined.")
        format = False
        for key, file in self.files.items(resolve=False):  # type: ignore
            # skip files we never touched
            if isinstance(file, LazyFile):
                logger.info("Skipping lazily unparsed file")
                continue
            file.save(self.serializer)
            if str(file.location).endswith(".tf"):
                format = True
        if format and self.terraform:
            try:
                self.terraform.run(["fmt"], path=dirname(self.path))
            except Exception as e:
                raise e
        if apply:
            self.apply()

    def save_all(self, format=True):
        self.save(format)
        for ws in self.children:
            ws.save_all(format)

    def find(self, object_type, invert=False):
        from pyterraformer.core.namespace import LazyFile

        output = defaultdict(list)
        replace = {}
        # resolve any files, so we can search them
        for key, file in self.files.items():
            if isinstance(file, LazyFile):
                replace[key] = file.resolve()
        for key, item in replace.items():
            self.files[key] = item
        for key, file in self.files.items():
            for object in file.objects:
                if isinstance(object, object_type):
                    output[key].append(object)
        if invert:
            out_list = {}
            for key, items in output.items():
                for item in items:
                    out_list[item] = key
            return out_list
        return output

    def get_object(self, **kwargs):
        for key, file in self.files.items():
            try:
                return file.get_object(**kwargs)
            except ValueError:
                pass
        raise ValueError(
            f"No object matching filter criteria {kwargs} found in any files in {self.path}"
        )


def value_match(item: Any, value: Any) -> bool:
    if isinstance(value, str):
        # first check for string comparison
        if item == value:
            return True
        # then fnmatch with quotes replaced
        return fnmatch(str(item).replace('"', ""), value)
    else:
        return item == value



import atexit
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from typing import Optional, Dict
from pyterraformer.core.generics import Backend
from pyterraformer.terraform.backends.base_backend import BaseBackend


@dataclass
class LocalBackend(BaseBackend):
    path: Optional[str] = None
    workspace_dir: Optional[str] = None

    def generate_environment(self) -> Dict:
        output: Dict = {}
        return output

    def as_object(self) -> Backend:
        return Backend(name="local")


@dataclass
class TemporaryLocalBackend(LocalBackend):
    def __init__(self):
        temp_dir = TemporaryDirectory()
        super().__init__(path=temp_dir)
        atexit.register(lambda: temp_dir.cleanup())



from .backend import Backend
from .comment import Comment
from .data import Data
from .interpolation import (
    Interpolation,
    DictLookup,
    PropertyLookup,
    StringLit,
    String,
    Expression,
    Conditional,
    BinaryOp,
    BinaryOperator,
    BinaryTerm,
    Parenthetical,
    File,
    Boolean,
    Merge,
    Concat,
    Replace,
    Types,
    ArrayLookup,
    GenericFunction,
    Symlink,
    LegacySplat,
    ToSet,
)
from .literal import Literal
from .local import Local
from .meta_arguments import DependsOn, Provider, ForEach, Count, Lifecycle
from .metadata import Metadata
from .output import Output
from .terraform import TerraformConfig
from .terraform_block import BlockList, BlockSet
from .variables import Variable

__all__ = [
    "Variable",
    "BlockList",
    "BlockSet",
    "TerraformConfig",
    "Output",
    "Local",
    "Literal",
    "Metadata",
    "DependsOn",
    "Provider",
    "ForEach",
    "Count",
    "Lifecycle",
    "Backend",
    "Comment",
    "Data",
    "Interpolation",
    "DictLookup",
    "PropertyLookup",
    "StringLit",
    "String",
    "Expression",
    "Conditional",
    "BinaryOp",
    "BinaryOperator",
    "BinaryTerm",
    "Parenthetical",
    "File",
    "Boolean",
    "Merge",
    "Concat",
    "Replace",
    "Types",
    "ArrayLookup",
    "GenericFunction",
    "Symlink",
    "LegacySplat",
    "ToSet",
]



import os
from pathlib import Path
from typing import Dict, List, Union, Optional, TYPE_CHECKING

from pyterraformer.enums import InsertPosition
from pyterraformer.serializer import BaseSerializer
from pyterraformer.core.utility import value_match

if TYPE_CHECKING:
    from pyterraformer.core.workspace import TerraformWorkspace
    from pyterraformer.core import TerraformObject


class TerraformNamespace(object):
    def __init__(
        self,
        name: str,
        workspace: "TerraformWorkspace",
        objects: Optional[List["TerraformObject"]] = None,
    ):
        self.workspace: "TerraformWorkspace" = workspace
        self.name = name
        self.objects: List = objects or []
        self.locals: Dict = {}

    def resolve(self):
        """No op for the base file"""
        return self


class LazyFile(TerraformNamespace):
    def __init__(self, file: Union[str, Path], workspace: "TerraformWorkspace"):
        super().__init__(
            name=os.path.basename(file).replace(".tf", ""), workspace=workspace
        )
        self.file = file

    def resolve(self) -> "TerraformFile":
        if not self.workspace.serializer:
            raise ValueError("No parser provided to look at files in this workspace")
        return self.workspace.serializer.parse_file(self.file, workspace=self.workspace)


class TerraformFile(TerraformNamespace):
    def __init__(
        self,
        workspace: "TerraformWorkspace",
        text: str,
        location: Union[str, Path],
        objects: Optional[List["TerraformObject"]] = None,
    ):
        self._text = text
        name = os.path.basename(location)
        super().__init__(name=name, workspace=workspace, objects=objects)
        self.location = location
        if self not in self.workspace.files:
            self.workspace.add_file(self)

    def get_object(self, **kwargs):
        for object in self.objects:
            if all(
                [
                    value_match(getattr(object, key, None), val)
                    for key, val in kwargs.items()
                ]
            ):
                return object
        raise ValueError(
            f"No object matching filter criteria {kwargs} found in file {self.location}"
        )

    def delete_object(self, object):
        orig = self.objects
        self.objects = [obj for obj in self.objects if obj != object]
        self.changed = orig != self.objects

    def find(self, object_type, invert=False):
        output = []
        for object in self.objects:
            if isinstance(object, object_type):
                output.append(object)
        if invert:
            return reversed(output)
        return output

    def _detect_duplicates(self, object) -> List:
        from pyterraformer.core.resources import ResourceObject
        from pyterraformer.core.modules import ModuleObject
        from pyterraformer.core.generics import Variable, Data

        duplicates = []
        if isinstance(object, (Variable, Data)):
            duplicates = [
                idx
                for idx, obj in enumerate(self.objects)
                if isinstance(object, (Variable, Data))
                and object.name == getattr(obj, "name", "")
            ]
        elif isinstance(object, ResourceObject):

            duplicates = [
                idx
                for idx, obj in enumerate(self.objects)
                if isinstance(object, ResourceObject)
                and object.tf_id
                and object.tf_id + (object._type or "")
                == getattr(obj, "tf_id", "") + getattr(obj, "_type", "")
            ]
        elif isinstance(object, ModuleObject):
            duplicates = [
                idx
                for idx, obj in enumerate(self.objects)
                if isinstance(object, ModuleObject)
                and object.tf_id
                and object.tf_id + (object._type or "")
                == getattr(obj, "tf_id", "") + getattr(obj, "_type", "")
            ]
        return duplicates

    def add_object(
        self,
        object: "TerraformObject",
        position: Union[InsertPosition, int] = InsertPosition.DEFAULT,
        exists_okay: bool = False,
        replace: bool = False,
    ):
        object._file = self
        duplicates = self._detect_duplicates(object)
        if duplicates:
            if replace:
                self.objects = [
                    obj
                    for idx, obj in enumerate(self.objects)
                    if idx not in (duplicates)
                ]
            elif exists_okay:
                return
            else:
                raise ValueError(
                    f"Duplicate resource name or ID detected {[obj for idx, obj in enumerate(self.objects) if idx in (duplicates)]} in file {self.name}! Cannot add unless the 'replace' or 'exists_okay' flags are set."
                )

        if position == InsertPosition.FIRST:
            self.objects.insert(0, object)
        elif position == InsertPosition.LAST:
            self.objects.append(object)
        elif position == InsertPosition.DEFAULT:
            indexes = [
                idx for idx, val in enumerate(self.objects) if val._type == object._type
            ]
            if indexes:
                self.objects.insert(indexes[-1] + 1, object)
            else:
                self.objects.append(object)
        elif isinstance(position, int):
            self.objects.insert(position, object)
        else:
            raise ValueError(f"Invalid Position Argument {position}")

    def render(self, serializer: BaseSerializer):
        return serializer.render_namespace(self)

    def save(self, serializer: BaseSerializer):
        try:
            with open(self.location, "w") as file:
                file.write(self.render(serializer))
        # if directory doesn't exist, create it
        except FileNotFoundError:
            import os

            os.makedirs(os.path.dirname(self.location))
            self.save(serializer=serializer)

    def __iter__(self):
        self._idx = 0
        return self

    def __next__(self):
        try:
            out = self.objects[self._idx]
            self._idx += 1
            return out

        except IndexError:
            self._idx = 0
            raise StopIteration  # Done iterating.



"""This module contains classes for resolving values out
of a parsed set of Terraform files.

These can have complex and condition resolution orders.

For example, consider the following
project = "${var.bigquery_processing_project["${terraform.workspace}"].test}"

This can be broken into the following

StringLit
    Interpolation
        PropertyLookup -> Resolve Left to Right
            DictLookup -> Resolve Left to Right
                StringLit
                    Interpolation
                        PropertyLookup -> Resolve Left to Right
            PropertyLookup -> Resolve Left to Right
"""

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from pyterraformer.core.namespace import TerraformFile
    from pyterraformer.core.workspace import TerraformWorkspace


@dataclass
class UnresolvedLookup:
    base: Any
    property: Any


class Resolvable:
    def resolve(self, workspace, file, parent, parent_instance):
        return self


def _resolver_function(arg, workspace, file, parent, parent_instance):
    if isinstance(arg, Resolvable):
        resolved = arg.resolve(workspace, file, parent, parent_instance)
        if resolved is None:
            raise ValueError("Unable to resolve function")
        return resolved
    else:
        if not arg:
            raise ValueError
        return arg


def variable_helper(arg, workspace, file, parent, parent_instance):
    from pyterraformer.core.generics.variables import Variable

    if isinstance(arg, Variable):
        return arg.default.resolve(workspace, file, parent, parent_instance)
    return arg


class FileLookupInstantiator(Resolvable):
    def __init__(self, workspace: "TerraformWorkspace"):
        for key, file in workspace.files.items():
            for object in file.objects:
                if hasattr(object, "tf_id"):
                    setattr(self, object.tf_id, object)
                if hasattr(object, "_name"):
                    setattr(self, object._name, object)


class FileObjectLookupInstantiator(Resolvable):
    def __init__(self, workspace: "TerraformWorkspace"):
        # TODO: 2022-06-05 figure out how to do this bette
        from pyterraformer.core.namespace import LazyFile

        self.workspace = workspace
        objects = []
        for key, file in self.workspace.files.items():
            if isinstance(file, LazyFile):
                file = file.resolve()
                workspace.files[key] = file

            for object in file.objects:
                objects.append(object)
        self.objects = objects

    def __getattr__(self, item):
        return FileObjectSubClassLookupInstantiator(
            [
                obj
                for obj in self.objects
                if hasattr(obj, "_type") and obj._type == str(item)
            ]
        )


class FileObjectSubClassLookupInstantiator(Resolvable):
    def __init__(self, obj_list):
        self.obj_list = obj_list

    def __getattr__(self, item):
        try:
            return [obj for obj in self.obj_list if obj.tf_id == item][0]
        except IndexError:
            return f"<could not compute {item} from file lookup>"


class DataLookupInstantiator(Resolvable):
    def __init__(self, workspace: "TerraformWorkspace"):
        self.workspace = workspace

    def __getattr__(self, item):
        return DataSubClassLookupInstantiator(
            [
                obj
                for obj in self.workspace.data
                if hasattr(obj, "type") and obj.type == str(item)
            ]
        )


class DataSubClassLookupInstantiator(Resolvable):
    def __init__(self, obj_list):
        self.obj_list = obj_list

    def __getattr__(self, item):
        return [obj for obj in self.obj_list if obj.name == item][0]


class VariableLookupInstantiator(Resolvable):
    def __init__(self, workspace: "TerraformWorkspace"):
        for key, value in workspace.variables.items():
            setattr(self, key, value)


class TerraformLookupInstantiator(Resolvable):
    def __init__(self, workspace: "TerraformWorkspace"):
        if workspace.terraform:
            self.workspace = workspace.terraform.workspace
        else:
            self.workspace = "undefined"


class LocalLookupInstantiator(Resolvable):
    def __init__(self, file: "TerraformFile"):
        for key, value in file.locals.items():
            setattr(self, key, value)


class CountLookupInstantiator(Resolvable):
    def __init__(self):
        pass

    @property
    def index(self):
        return 1


class Interpolation(Resolvable):
    def __init__(self, contents):
        self.contents = contents

    def __repr__(self):
        return "${{{}}}".format("".join([f"{val.__repr__()}" for val in self.contents]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        parent = None
        # an interpolation always resets parent_instances
        parent_instance = self
        resolve = deepcopy(self.contents)
        resolved = []
        while resolve:
            next = resolve.pop(0)
            parent = _resolver_function(next, workspace, file, parent, parent_instance)
            resolved.append(parent)
            parent_instance = next
        if parent is None:
            raise ValueError(f"No values for interpolation {resolved}")
        return parent


class DictLookup(Resolvable):
    def __init__(self, base, lookup):
        self.base = base
        # TODO don't return a list here
        self.contents = lookup
        # the lookup will be a nested list
        self.lookup = lookup[0]

    def __repr__(self):
        return f"{self.base.__repr__()}[{self.lookup.__repr__()}]"

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        """ex: bigquery_processing_project["${terraform.workspace}"]
        to resolve a dict lookup
        we will resolve the lookup, then
        """
        parent = parent or self.base
        lookup = deepcopy(self.lookup)
        lookup = lookup.resolve(workspace, file, None, self)
        results = parent[lookup]
        return _resolver_function(results, workspace, file, None, self)


class ArrayLookup(Resolvable):
    def __init__(self, base, lookup):
        self.base = base
        # TODO don't return a list here
        self.contents = lookup
        # the lookup will be a nested list
        self.lookup = lookup[0]

    def __repr__(self):
        return f"{self.base.__repr__()}[{self.lookup.__repr__()}]"

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        """ex: bigquery_processing_project["${terraform.workspace}"]
        to resolve a dict lookup
        we will resolve the lookup, then
        """
        parent = parent or self.base
        lookup = deepcopy(self.lookup)
        lookup = int(lookup.resolve(workspace, file, None, self))
        results = parent[lookup]
        return _resolver_function(results, workspace, file, None, self)


class PropertyLookup(Resolvable):
    def __init__(self, base, attributes: List):
        self.base = base
        self.contents = attributes
        self.property = attributes[0]

    def __repr__(self):
        return "{}.{}".format(self.base.__repr__(), self.property.__repr__())

    def resolve(
        self,
        workspace: "TerraformWorkspace",
        file: "TerraformFile",
        parent: Optional[Resolvable] = None,
        parent_instance: Optional[Resolvable] = None,
    ):

        anchor = parent or self.base
        if self.base == "module":
            anchor = FileLookupInstantiator(workspace)
        elif self.base == "var":
            anchor = VariableLookupInstantiator(workspace)
        elif self.base == "terraform":
            anchor = TerraformLookupInstantiator(workspace)
        elif self.base == "data":
            anchor = DataLookupInstantiator(workspace)
        elif self.base == "locals":
            anchor = LocalLookupInstantiator(file)
        elif self.base == "count":
            anchor = CountLookupInstantiator()
        elif not parent:
            anchor = getattr(FileObjectLookupInstantiator(workspace), str(self.base))
        # the lookup could be another property or a dictionary
        # in either case, pull out the base value and look that up immediately
        # then pass forward

        if isinstance(self.property, (DictLookup, PropertyLookup)):
            local_resolved = getattr(anchor, str(self.property.base))
            return self.property.resolve(
                workspace, file, parent=local_resolved, parent_instance=self
            )
        else:
            try:
                return getattr(anchor, str(self.property))
            except AttributeError:
                return UnresolvedLookup(anchor, self.property)


class StringLit(Resolvable):
    def __init__(self, contents: List):
        self.contents = contents

    @property
    def string(self) -> str:
        return "".join([str(v) for v in self.contents])

    def __add__(self, v):
        return self.string.__add__(v)

    def __hash__(self):
        return hash(str(self))

    def split(self, splitter: str = " "):
        return str(self.contents).split(splitter)

    def __eq__(self, other):
        return str(self).replace('"', "") == str(other) or str(self) == str(other)

    def __repr__(self):
        return '"{}"'.format("".join([val.__repr__() for val in self.contents]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        return "".join(
            [
                str(
                    variable_helper(
                        item.resolve(workspace, file, parent, parent_instance),
                        workspace,
                        file,
                        parent,
                        parent_instance,
                    )
                )
                for item in self.contents
            ]
        )


class String(Resolvable):
    def __init__(self, item):
        self.item = item

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return self.item

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        if isinstance(parent_instance, PropertyLookup):
            resolved = PropertyLookup(self.item).resolve(
                workspace, file, parent, parent_instance
            )
            return resolved
        return self.item


class File(Resolvable):
    def __init__(self, item):
        self.item = item[0] if isinstance(item, list) else item

    def __repr__(self):
        return "file({})".format(self.item.__repr__())

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        if isinstance(parent_instance, PropertyLookup):
            resolved = PropertyLookup(self.item).resolve(
                workspace, file, parent, parent_instance
            )
            out = resolved
        else:
            out = self.item
        return f"file({out})"


class Concat(Resolvable):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return "concat({})".format(",".join([item.__repr__() for item in self.items]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        final = []
        for item in self.items:
            if isinstance(parent_instance, PropertyLookup):
                resolved = PropertyLookup(item).resolve(
                    workspace, file, parent, parent_instance
                )
                out = resolved
            elif isinstance(item, Resolvable):
                out = item.resolve(workspace, file, None, None)
            else:
                out = item
            final.append(out)
        if all([isinstance(item, str) for item in final]):
            concat = "".join(final)
        else:
            concat = []
            for item in final:
                concat += item
        return f"concat({concat})"


class Types(Resolvable):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return "types({})".format(",".join([item.__repr__() for item in self.items]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        # final = []
        # for item in self.items:
        #     if isinstance(parent_instance, PropertyLookup):
        #         resolved = PropertyLookup(item).resolve(workspace, file, parent, parent_instance)
        #         out = resolved
        #     else:
        #         out = item
        #     final.append(out)
        # concat = ",".join(final)
        return "types({})".format(",".join([item.__repr__() for item in self.items]))


class Replace(Resolvable):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return "replace({})".format(",".join([item.__repr__() for item in self.items]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        final = []
        for item in self.items:
            if isinstance(parent_instance, PropertyLookup):
                resolved = PropertyLookup(item).resolve(
                    workspace, file, parent, parent_instance
                )
                out = resolved
            else:
                out = item
            final.append(out)
        concat = ",".join(final)
        return f"replace({concat})"


class GenericFunction(Resolvable):
    def __init__(self, items):
        self.name = str(items[0])
        self.items = items[1:]

    def __repr__(self):
        return "{}({})".format(
            self.name, ",".join([item.__repr__() for item in self.items])
        )

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        final = []
        for item in self.items:
            if isinstance(parent_instance, PropertyLookup):
                resolved = PropertyLookup(item).resolve(
                    workspace, file, parent, parent_instance
                )
                out = resolved
            else:
                out = item
            final.append(out)
        concat = ",".join([item.__repr__() for item in final])
        return f"{self.name}({concat})"


class Merge(Resolvable):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return "merge({})".format(",".join([item.__repr__() for item in self.items]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        final = []
        for item in self.items:
            if isinstance(parent_instance, PropertyLookup):
                resolved = PropertyLookup(item).resolve(
                    workspace, file, parent, parent_instance
                )
                out = resolved
            else:
                out = item
            final.append(out)
        concat = ",".join(final)
        return f"merge({concat})"


class Conditional(Resolvable):
    def __init__(self, args):
        self.bool = args[0]
        self.true = args[1]
        self.false = args[2]

    def __repr__(self):
        return f"{self.bool} ? {self.true} : {self.false}"

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        boolean_eval = self.bool.resolve(workspace, file, parent, parent_instance)
        if boolean_eval:
            return self.true.resolve(workspace, file, parent, parent_instance)
        else:
            return self.false.resolve(workspace, file, parent, parent_instance)


class Parenthetical(Resolvable):
    def __init__(self, args):
        self.contents = args

    def __repr__(self):
        return "({})".format("".join([val.__repr__() for val in self.contents]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        # parent = None
        parent_instance = self
        resolve = deepcopy(self.contents)
        while resolve:
            next = resolve.pop(0)
            parent = _resolver_function(next, workspace, file, parent, parent_instance)
            parent_instance = next
        if parent is None:
            raise ValueError(f"No values for interpolation {self.contents}")
        return parent


class Expression(Resolvable):
    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return "".join([val.__repr__() for val in self.args])

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        output = self.args[0]
        if hasattr(output, "resolve"):
            return self.args[0].resolve(workspace, file, parent, parent_instance)
        else:
            return output


class BinaryOp(Resolvable):
    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return "".join([val.__repr__() for val in self.args])

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        all = [item.resolve(workspace, file, parent) for item in self.args]
        left = all[0]
        operator, right = all[1]
        if isinstance(parent_instance, PropertyLookup):
            left = PropertyLookup(left).resolve(
                workspace, file, parent, parent_instance
            )
        if operator == "==":
            return left == right
        elif operator == ">=":
            return left >= right
        elif operator == "<=":
            return left <= right
        elif operator == "!=":
            return left != right
        elif operator == "%":
            return True
        else:
            raise ValueError(f"Unable to do comparison {left}, {operator}, {right}")


class BinaryOperator(Resolvable):
    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return str(self.args[0].value)

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        return str(self.args[0].value)


class BinaryTerm(Resolvable):
    def __init__(self, args):
        self.args = args

    def __repr__(self):
        return "".join([val.__repr__() for val in self.args])

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        return [
            item.resolve(workspace, file, parent, parent_instance) for item in self.args
        ]


class Boolean(Resolvable):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value.__eq__(other)

    def __repr__(self):
        if self.value:
            return "true"
        return "false"


class Block(Resolvable):
    log_keys = False

    def __init__(self, name, values):
        self._keys = {}
        self.name = str(name)
        self.values = values
        for attribute in self.values:
            if isinstance(attribute, list):
                # always cast keys to string
                key = str(attribute[0])
                val = attribute[1]
                if isinstance(val, Block):
                    base = getattr(self, key, Block)
                    base.append(val)
                    val = base
                setattr(self, key, val)
                self._keys[key] = val
            elif isinstance(attribute, Block):
                key = str(attribute.name)
                base = getattr(self, attribute.name, Block())
                base.append(attribute)
                setattr(self, key, base)
                self._keys[key] = attribute
            self.log_keys = True

    def __setattr__(self, key, value):
        if self.log_keys and key in self._keys:
            self._keys[key] = value
        super().__setattr__(key, value)

    def __repr__(self):
        out = []
        for key, item in self._keys.items():
            if isinstance(item, Block):
                for sub in item:
                    out.append(f"{key} {sub.__repr__()}")
            elif isinstance(item, Block):
                out.append(f"{key} {item.__repr__()}")
            else:
                out.append(f"{key}={item.__repr__()}")
        out = "\n".join(out)
        return f"""{{
        {out}
        }}"""


class Symlink(Resolvable):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value.__eq__(other)

    def __repr__(self):
        return "".join([val.__repr__() for val in self.value])


class LegacySplat(Resolvable):
    def __init__(self, args):
        self.contents = args

    def __repr__(self):
        return "{}".format(*[val.__repr__() for val in self.contents[:1]])

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        # parent = None
        parent_instance = self
        resolve = deepcopy(self.contents)
        while resolve:
            next = resolve.pop(0)
            parent = _resolver_function(next, workspace, file, parent, parent_instance)
            parent_instance = next
        if parent is None:
            raise ValueError(f"No values for interpolation {self.contents}")
        return parent


class ToSet(Resolvable):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return "toset({})".format(",".join([item.__repr__() for item in self.items]))

    def resolve(self, workspace, file, parent=None, parent_instance=None):
        final = []
        for item in self.items:
            if isinstance(parent_instance, PropertyLookup):
                resolved = PropertyLookup(parent_instance, item).resolve(
                    workspace, file, parent, parent_instance
                )
                out = resolved
            else:
                out = item
            final.append(out)
        concat = ",".join(final)
        return f"toset({concat})"



from pathlib import Path
import os
from logging import DEBUG
from logging import StreamHandler

from pyterraformer import HumanSerializer
from pyterraformer.constants import logger
from pyterraformer.core import TerraformWorkspace
from pyterraformer.terraform import Terraform
from pyterraformer.terraform.backends.local_backend import LocalBackend

logger.addHandler(StreamHandler())
logger.setLevel(DEBUG)


def test_all_parsing():
    test_cases = Path(__file__).parent / "cases"
    tf = Terraform(terraform_exec_path=None, backend=LocalBackend(path=test_cases))
    workspace = TerraformWorkspace(
        terraform=tf, path=test_cases, serializer=HumanSerializer(terraform=tf)
    )

    all_files = os.listdir(test_cases)
    for file in all_files:
        filepath = Path(file)
        namespace = workspace.get_file_safe(filepath.name)
        assert len(namespace.objects) > 0



from pyterraformer.core.resources.resource_object import ResourceObject


class GenericResource(ResourceObject):
    """This resource represents an object that is parsed
    from a terraform file but there is no matching provider library
    to provide type hinting and auto completion"""

    def __init__(self, tf_id, rtype, text, attributes):
        self._type = str(rtype).replace('"', "")
        ResourceObject.__init__(self, rtype, tf_id, text, attributes)



from typing import List

from pyterraformer.core.generics import Literal
from pyterraformer.core.objects import TerraformObject


class BaseModule(TerraformObject):
    REQUIRED_ATTRIBUTES: List[str] = []
    PRIORITY_ATTRIBUTES: List[str] = []

    def __init__(self, text, name, attributes):
        self._name = str(name).replace('"', "")
        self.id = self._name
        TerraformObject.__init__(self, "module", text, attributes)

    def render_attribute(self, item):
        return f"${{module.{self._name}.{item}}}"

    @property
    def filtered_source(self):
        return str(self.source).replace('"', "")

    @property
    def short_name(self):
        if self.filtered_source.startswith("http"):
            return extract_http_module(str(getattr(self, "source", None)))
        elif self.filtered_source.startswith("git"):
            return extract_git_module(str(getattr(self, "source", None)))
        else:
            raise ValueError(f"Unable to parse source {self.source}")

    @property
    def version(self):
        if self.filterered_source.startswith("http"):
            return extract_http_version(self.filterered_source)
        # elif self.filtered_source.startswith('git'):
        #     return extract_http_version(self.filterered_source)
        else:
            raise ValueError(f"Unable to parse version from source {self.source}")

    @classmethod
    def create(cls, id, **kwargs):
        kwargs = kwargs or {}
        attributes = [[key, item] for key, item in kwargs.items()]
        if hasattr(cls, "REQUIRED_ATTRIBUTES"):
            for attribute in cls.REQUIRED_ATTRIBUTES:
                if attribute not in kwargs:
                    raise ValueError(f"Missing required attribute {attribute}")
        return cls(name=id, text=None, attributes=attributes)

    def interpolation_property(self, property: str):
        from pyterraformer.core.generics import Literal

        return Literal(f"${{module.{self.id}.{property}}}")

    def __getattr__(self, item):
        if item.endswith("_lookup"):
            property = item.split("_lookup")[0]
            return Literal(f"{self._type}.{self.id}.{property}")
        return super().__getattr__(item)


def extract_git_module(input: str):
    return input.split("/")[-1].split(".")[0]


def extract_http_module(input: str):
    return input.split("/")[-1].split("_")[0]


def extract_http_version(input: str):
    return ".".join(input.split("/")[-1].split("_")[1].split(".")[0:3])[1:]



try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

from .config import Config
from .core.workspace import TerraformWorkspace
from .serializer import HumanSerializer
from .terraform.backends import LocalBackend
from .terraform.terraform import Terraform

try:
    __version__ = version("pyterraformer")
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = [
    "HumanSerializer",
    "Config",
    "Terraform",
    "LocalBackend",
    "TerraformWorkspace",
]



from pyterraformer import HumanSerializer
from pyterraformer.core import TerraformObject


def standard_string(string: str):
    return " ".join(string.split())


def test_human_serialization():
    hs = HumanSerializer()
    example_string = """resource "aws_s3_bucket" "b" {
      bucket = "my-tf-test-bucket"
    
      tags = {
        Name        = "My bucket"
        Environment = "Dev"
      }
    }
    
    resource "aws_s3_bucket" "b2" {
      bucket = "my-tf-test-bucket"
    
      tags = {
        Name        = "My bucket"
        Environment = "Dev"
      }
    }"""
    bucket = hs.parse_string(example_string)[0]

    bucket.tags["Environment"] = "Prod"
    bucket.bucket = "my-updated-bucket"

    test_string = """resource "aws_s3_bucket" "b" {
      bucket = "my-updated-bucket"
    
      tags = {
        Name        = "My bucket"
        Environment = "Prod"
      }
    }"""

    updated = hs.render_object(bucket)

    assert standard_string(updated) == standard_string(test_string)


def test_block_parsing(human_serializer):

    split = """terraform {
  backend "local" {

  }
  required_providers {
    test = {
      source = "hashicorp/google"
      version = "4.27.0"
    }
  }
  required_providers {
    build = {
      source  = "hashicorp/google"
      version = "4.27.0"
    }
  }
}

"""
    parsed_split = human_serializer.parse_string(split)[0]
    assert len(parsed_split.required_providers) == 2

    joined = """terraform {
  backend "local" {

  }
  required_providers {
    test = {
      source = "hashicorp/google"
      version = "4.27.0"
    }
    build = {
      source  = "hashicorp/google"
      version = "4.27.0"
    }
  }
}

"""
    parsed_joined = human_serializer.parse_string(joined)[0]

    assert len(parsed_joined.required_providers) == 2


def test_block_parsing_variants(human_serializer):

    split = """terraform {
  backend "local" {

  }
  required_providers {
    test = {
      source = "hashicorp/google"
      version = "4.27.0"
    }
  }
  required_providers {
    build = {
      source  = "hashicorp/google"
      version = "4.27.0"
    }
  }
}

"""

    joined = """terraform {
  backend "local" {

  }
  required_providers {
    test = {
      source = "hashicorp/google"
      version = "4.27.0"
    }
    build = {
      source  = "hashicorp/google"
      version = "4.27.0"
    }
  }
}

"""
    parsed_split = human_serializer.parse_string(split)[0]
    parsed_joined = human_serializer.parse_string(joined)[0]

    assert parsed_split.required_providers == parsed_joined.required_providers


def test_comments(human_serializer):

    split = """resource "google_container_cluster" "primary" {
  name     = "my-gke-cluster"
  location = "us-central1"

  # We can't create a cluster with no node pool defined, but we want to only use
  # separately managed node pools. So we create the smallest possible default
  # node pool and immediately delete it.
  remove_default_node_pool = true
  initial_node_count       = 1
}
"""
    parsed_split = human_serializer.parse_string(split)[0]

    rendered = human_serializer.render_object(parsed_split)
    print(rendered)
    assert (
        """# We can't create a cluster with no node pool defined, but we want to only use
# separately managed node pools. So we create the smallest possible default
# node pool and immediately delete it."""
        in rendered
    )


def test_comments_reserialization(human_serializer):
    base = """resource "aws_s3_bucket" "b" {
  test = false
  # maintain position
  bucket = "my-tf-test-bucket"
  # this is a helpful comment
  tags = {
    Name        = "My bucket"
    Environment = "Dev"
  }
}
# this is a top level comment

/*
MULTILINE
*/
"""

    parsed_split = human_serializer.parse_string(base)
    assert len(parsed_split) == 3
    for object in parsed_split:
        assert isinstance(object, TerraformObject)
    resource = parsed_split[0]

    rendered = human_serializer.render_object(resource)
    assert "# maintain position" in rendered
    assert "# this is a helpful comment" in rendered
    assert rendered.find("# maintain position") < rendered.find(
        "# this is a helpful comment"
    )



from .namespace import TerraformNamespace, TerraformFile
from .objects import TerraformObject
from .workspace import TerraformWorkspace

__all__ = [
    "TerraformObject",
    "TerraformWorkspace",
    "TerraformFile",
    "TerraformNamespace",
]



from os.path import join, dirname
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory
from typing import Optional, Dict, Union, TYPE_CHECKING, Any, List
from dataclasses import is_dataclass, asdict


import jinja2

from pyterraformer.constants import logger, EMPTY_DEFAULT
from pyterraformer.core.generics import Backend, Comment
from pyterraformer.core.modules import ModuleObject
from pyterraformer.core.resources import ResourceObject
from pyterraformer.serializer.base_serializer import BaseSerializer
from pyterraformer.serializer.human_resources.engine import parse_text
from pyterraformer.exceptions import TerraformExecutionError

if TYPE_CHECKING:
    from pyterraformer.terraform import Terraform
    from pyterraformer.core import (
        TerraformWorkspace,
        TerraformObject,
        TerraformNamespace,
        TerraformFile,
    )

TEMPLATE_PATH = join(dirname(__file__), "templates")

template_loader = jinja2.FileSystemLoader(searchpath=TEMPLATE_PATH)
env = jinja2.Environment(
    loader=template_loader, autoescape=True, keep_trailing_newline=True
)


def process_attribute(input: Any):
    from pyterraformer.core.generics import Variable, Literal, BlockList, BlockSet

    valid = False
    if isinstance(input, dict) or is_dataclass(input):
        valid = True
    if not valid:
        return input

    if is_dataclass(input):
        final_input = asdict(input)
    else:
        final_input = input
    output: Dict[str, Any] = {}
    for key, item in final_input.items():
        if item == EMPTY_DEFAULT:
            continue
        elif isinstance(item, Variable):
            output[key] = item.render_basic()
        elif isinstance(item, Literal):
            output[key] = item
        elif str(key).startswith("comment-") and isinstance(item, Comment):
            output[key] = Literal(item.text)
        elif isinstance(item, (BlockList, BlockSet)):
            for idx, sub_item in enumerate(item):  # type: ignore
                output[f"{key}~~block_{idx}"] = process_attribute(sub_item)
        elif is_dataclass(item):
            output[f"{key}~~block_0"] = process_attribute(asdict(item))
        elif isinstance(item, Backend):
            output[f"{key}~~block_0"] = process_attribute(item)
        elif isinstance(item, dict):
            output[key] = process_attribute(item)
        elif isinstance(item, List):
            output[key] = [process_attribute(sub_item) for sub_item in item]
        else:
            output[key] = item
    return output


class HumanSerializer(BaseSerializer):
    def __init__(self, terraform: Optional[Union[str, "Terraform"]] = None):
        from pyterraformer.terraform import Terraform

        self.terraform: Optional[Terraform] = None
        if isinstance(terraform, Terraform):
            self.terraform = terraform
        elif terraform:
            self.terraform = Terraform(terraform_exec_path=terraform)

    #
    # def _format_path(self):
    #     self.terraform.

    @property
    def can_format(self) -> bool:
        return self.terraform is not None

    def parse_string(self, string: str):
        return parse_text(string)

    def parse_file(self, path: Union[str, Path], workspace: "TerraformWorkspace"):
        from pyterraformer.core.namespace import TerraformFile

        with open(path, "r") as f:
            text = f.read()
            objects = self.parse_string(string=text)
        return TerraformFile(
            location=path, workspace=workspace, objects=objects, text=text
        )

    def _format_string(self, string: str) -> str:
        if not self.terraform:
            return string
        with TemporaryDirectory() as td:
            temp_dir = Path(td)
            file_name = temp_dir / "to_format.tf"
            file_name.write_text(string)
            try:
                self.terraform.run("fmt", path=td)
            except FileNotFoundError as e:
                logger.error(str(e))
                raise TerraformExecutionError(
                    f"File not found - is the terraform executable path set correctly and accessible to this user? Error: {str(e)}"
                )
            except CalledProcessError as e:
                logger.error(f"Unable to format file \n{string}")
                raise e
            return file_name.open().read()

    def render_object(
        self, object: "TerraformObject", format: Optional[bool] = None
    ) -> str:
        if format and not self.can_format:
            raise ValueError("No terraform executable configured, cannot format.")
        from pyterraformer.core.generics import TerraformConfig

        format = format if format is not None else self.can_format
        variables = {}
        variables["tf_id"] = object.tf_id
        variables["type"] = object._type
        final = {}
        for key in object.render_variables.keys():
            if key not in final:
                final[key] = object.render_variables[key]
        if isinstance(object, TerraformConfig):
            template_name = "terraform.tf"
        elif isinstance(object, ResourceObject):
            template_name = "resource.tf"
        elif isinstance(object, ModuleObject):
            template_name = "module.tf"
        else:
            template_name = "generic.tf"
        template = env.get_template(template_name)
        # print(final)
        # print(process_attribute(final))
        # raise ValueError
        string = template.render(
            render_attributes=process_attribute(final), **variables
        )

        if format:
            string = self._format_string(string)
        return string

    def render_namespace(
        self, namespace: "TerraformNamespace", format: Optional[bool] = None
    ) -> str:
        from pyterraformer.core.generics import Comment

        format = format if format is not None else self.can_format

        out = []
        for idx, object in enumerate(namespace.objects):
            # comments should have no trailing whitespace
            if isinstance(object, Comment):
                out.append(self.render_object(object, format=False))
            # EOF one
            elif idx == len(namespace.objects) - 1:
                out.append(self.render_object(object, format=False) + "\n")
            # all others two
            else:
                out.append(self.render_object(object, format=False) + "\n\n")
        if format:
            string = "".join(out)
            return self._format_string(string)
        return "".join(out)

    def render_workspace(
        self, workspace: "TerraformWorkspace", format: Optional[bool] = None
    ) -> Dict[str, str]:
        format = format if format is not None else self.can_format
        output = {}
        for name, file in workspace.files.items():
            output[name] = self.render_namespace(file, format=format)
        return output



from dataclasses import dataclass
from typing import Optional

from pyterraformer.settings import get_default_terraform_location


# tempdir = mkdtemp()
#
# atexit.register(shutil.rmtree, tempdir)


@dataclass
class TerraformerConfig(object):
    terraform_exec: Optional[str] = get_default_terraform_location()
    default_workspace: str = "default"
    tf_plugin_cache_dir: Optional[str] = None
    default_variable_file: str = "variables.tf"
    default_data_file: str = "data.tf"

    #
    # def configure_git_module_provider(self):
    #     pass

    @property
    def state_provider(self):
        return None


Config = TerraformerConfig()



import os
from fnmatch import fnmatch
from pathlib import Path
from typing import Iterator, List, Union


def splitall(path: str) -> Iterator[str]:
    out: List[str] = []
    while True:
        base, end = os.path.split(path)
        if base and end:
            out.append(end)
        else:
            out.append(base)
            break
        path = base
    return reversed(out)


def get_root(path: str, breaker: Union[str, List[str]] = None):
    if isinstance(breaker, str):
        breaker = [breaker]
    elif not breaker:
        breaker = ["terraform", "terraform-local-cache"]
    else:
        breaker = breaker
    parts = splitall(path)
    base = []
    for item in parts:
        base.append(item)
        if item.lower() in breaker:
            break
    joined = os.path.join(*base)
    return str(Path(os.path.relpath(path, joined)).as_posix())


def value_match(item, value) -> bool:
    if isinstance(value, str):
        # first check for string comparison
        if item == value:
            return True
        # then fnmatch with quotes replaced
        return fnmatch(str(item).replace('"', ""), value)
    else:
        return item == value



from pyterraformer.core.objects import TerraformObject


class Data(TerraformObject):
    _type = "data"

    def __init__(self, name, type, text, attributes):
        self.name = str(name).replace('"', "")
        self.type = str(type).replace('"', "")
        TerraformObject.__init__(self, self._type, text, attributes)

    def __repr__(self):
        return (
            f"{self._type}({self.name})("
            + ", ".join(
                [f'{key}="{val}"' for key, val in self.render_variables.items()]
            )
            + ")"
        )

    #
    # def render(self, variables=None):
    #     variables = variables or {}
    #     final = {}
    #     for key in sorted(self.render_variables.keys()):
    #         if key not in final and key != "type":
    #             final[key] = self.render_variables[key]
    #
    #     for key, item in final.items():
    #         if (
    #             isinstance(item, str)
    #             and item.startswith('"${')
    #             and item.endswith('$}"')
    #         ):
    #             item = item[3:-3]
    #             final[key] = Literal(item)
    #
    #     output = process_attribute(final)
    #     return self.template.render(
    #         name=self.name, type=self.type, render_attributes=output, **variables
    #     )



from dataclasses import dataclass
from typing import List, Optional

from pyterraformer.core.objects import TerraformObject, ObjectMetadata


@dataclass
class StateResponse:
    resource_name: str
    remote_address: str


class ResourceObject(TerraformObject):
    REQUIRED_ATTRIBUTES: List[str] = []
    PRIORITY_ATTRIBUTES: List[str] = []
    BLOCK_ATTRIBUTES: List[str] = []
    _type = "generic_resource_object"

    def __init__(
        self, tf_id: str, _metadata: Optional[ObjectMetadata] = None, **kwargs
    ):
        TerraformObject.__init__(self, self._type, tf_id, _metadata=_metadata, **kwargs)

    def render_attribute(self, item):
        return f"${{{self._type}.{self.tf_id}.{item}}}"

    @property
    def import_address(self) -> str:
        raise NotImplementedError



# from analytics_utility_core.decorators import lazy_property

from typing import Optional
from pyterraformer.core.objects import TerraformObject, ObjectMetadata


class Provider(TerraformObject):
    def __init__(self, type: str, _metadata: Optional[ObjectMetadata] = None, **kwargs):
        self.ptype = str(type).replace('"', "")
        TerraformObject.__init__(self, _type="provider", _metadata=_metadata)

    def __repr__(self):
        return (
            f"{self._type}-{self.ptype}("
            + ", ".join(
                [f'{key}="{val}"' for key, val in self.render_variables.items()]
            )
            + ")"
        )

    # def render(self, variables=None):
    #     # from analytics_terraformer_core.utility import clean_render_dictionary
    #
    #     variables = variables or {}
    #     variables["type"] = self.ptype
    #
    #     # final = clean_render_dictionary(self.render_variables, [])
    #     output = process_attribute(self.render_variables)
    #     return self.template.render(render_attributes=output, **variables)


class Lifecycle(TerraformObject):
    def __init__(self, text, attributes):
        TerraformObject.__init__(self, "lifecycle", text, attributes)


class Count(TerraformObject):
    def __init__(self, text, attributes):
        TerraformObject.__init__(self, "count", text, attributes)


class DependsOn(TerraformObject):
    def __init__(self, text, attributes):
        TerraformObject.__init__(self, "depends_on", text, attributes)


class ForEach(TerraformObject):
    def __init__(self, text, attributes):
        TerraformObject.__init__(self, "for_each", text, attributes)
